/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.microtesk.translator.mmu.spec;

import java.util.List;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.microtesk.translator.mmu.spec.basis.IntegerField;
import ru.ispras.microtesk.translator.mmu.spec.basis.IntegerVariable;

/**
 * This class describes an assignment.
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public class MmuAssignment {
  /** The field. */
  private final IntegerField field;
  /** The expression. */
  private final MmuExpression expression;

  /**
   * Constructs an assignment.
   * 
   * @param field the field.
   * @param expression the expression.
   * @throws NullPointerException if {@code field} is null.
   */
  public MmuAssignment(final IntegerField field, final MmuExpression expression) {
    InvariantChecks.checkNotNull(field);

    this.field = field;
    this.expression = expression;
  }

  /**
   * Constructs an assignment.
   * 
   * @param variable the variable.
   * @param expression the expression.
   * @throws NullPointerException if {@code variable} is null.
   */
  public MmuAssignment(final IntegerVariable variable, final MmuExpression expression) {
    this(new IntegerField(variable), expression);
  }

  /**
   * Constructs an assignment with no right-hand side (RHS). It is assumed that RHS can be derived
   * from context.
   * 
   * @param field the field.
   */
  public MmuAssignment(final IntegerField field) {
    this(field, null);
  }

  /**
   * Constructs an assignment with no right-hand side (RHS). It is assumed that RHS can be derived
   * from context.
   * 
   * @param variable the variable.
   */
  public MmuAssignment(final IntegerVariable variable) {
    this(new IntegerField(variable), null);
  }

  /**
   * Returns the field of the assignment.
   * 
   * @return the field.
   */
  public IntegerField getField() {
    return field;
  }

  /**
   * Returns the expression of the assignment.
   * 
   * @return the expression.
   */
  public MmuExpression getExpression() {
    return expression;
  }

  @Override
  public String toString() {
    final StringBuilder string = new StringBuilder(field.getVariable().getName());

    string.append("[");
    string.append(field.getLoIndex());
    string.append(":");
    string.append(field.getHiIndex());
    string.append("]: ");

    if (expression == null) {
      string.append("{null}");
    } else {
      final List<IntegerField> terms = expression.getTerms();
      for (final IntegerField term : terms) {
        string.append("{");
        string.append(term);
        string.append("}");
      }
    }

    return string.toString();
  }
}
